/* umockdev-utils.h generated by valac 0.56.19, the Vala compiler, do not modify */

#ifndef __UMOCKDEV_UTILS_H__
#define __UMOCKDEV_UTILS_H__

#include <stdlib.h>
#include <string.h>
#include <glib.h>

G_BEGIN_DECLS

#if !defined(VALA_EXTERN)
#if defined(_MSC_VER)
#define VALA_EXTERN __declspec(dllexport) extern
#elif __GNUC__ >= 4
#define VALA_EXTERN __attribute__((visibility("default"))) extern
#else
#define VALA_EXTERN extern
#endif
#endif

VALA_EXTERN void umockdev_utils_checked_setenv (const gchar* variable,
                                    const gchar* value);
VALA_EXTERN void umockdev_utils_checked_remove (const gchar* path);
VALA_EXTERN void umockdev_utils_remove_dir (const gchar* path,
                                gboolean remove_toplevel);
VALA_EXTERN GPid umockdev_utils_spawn_process_under_test (gchar** argv,
                                              gint argv_length1,
                                              GChildWatchFunc watch_cb,
                                              gpointer watch_cb_target,
                                              GDestroyNotify watch_cb_target_destroy_notify,
                                              GError** error);

G_END_DECLS

#endif
